@@ -2,8 +2,8 @@
 /*
  This file is part of SLRN.
 
- Copyright (c) 1994, 1999 John E. Davis <davis@space.mit.edu>
- Copyright (c) 2001, 2002 Thomas Schultz <tststs@gmx.de>
+ Copyright (c) 1994, 1999, 2007-2012 John E. Davis <jed@jedsoft.org>
+ Copyright (c) 2001-2006 Thomas Schultz <tststs@gmx.de>
 
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
@@ -28,7 +28,6 @@
 #include <string.h>
 #include <time.h>
 
-
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
 #endif
@@ -44,17 +43,17 @@
 #include "slrn.h"
 #include "group.h"
 #include "util.h"
+#include "strutil.h"
 #include "server.h"
 #include "art.h"
 #include "misc.h"
 #include "post.h"
 /* #include "clientlib.h" */
 #include "startup.h"
-#include "hash.h" 
+#include "hash.h"
 #include "score.h"
 #include "menu.h"
 #include "xover.h"
-#include "chmap.h"
 #include "print.h"
 
 /*}}}*/
@@ -62,6 +61,7 @@
 static int Art_Hide_Quote_Level = 1;
 int Slrn_Wrap_Mode = 3;
 int Slrn_Wrap_Method = 2;
+int Slrn_Wrap_Width = -1;
 
 static char *Super_Cite_Regexp = "^[^A-Za-z0-9]*\"\\([-_a-zA-Z/]+\\)\" == .+";
 
@@ -71,7 +71,7 @@ static void hide_article_lines (Slrn_Article_Type *a, unsigned int mask) /*{{{*/
 
    if (a == NULL)
      return;
-   
+
    a->needs_sync = 1;
    l = a->lines;
 
@@ -107,7 +107,7 @@ static void unhide_article_lines (Slrn_Article_Type *a, unsigned int mask) /*{{{
 
 /*}}}*/
 
-/* The function that set the needs_sync flag article line number start with _ 
+/* The function that set the needs_sync flag article line number start with _
  */
 
 int _slrn_art_unhide_quotes (Slrn_Article_Type *a) /*{{{*/
@@ -122,22 +122,48 @@ int _slrn_art_unhide_quotes (Slrn_Article_Type *a) /*{{{*/
 
 /*}}}*/
 
+static unsigned char *is_matching_line (unsigned char *b, SLRegexp_Type **r) /*{{{*/
+{
+   unsigned int len;
+   len = strlen ((char *) b);
 
+   while (*r != NULL)
+     {
+	SLRegexp_Type *re;
+	unsigned int match_len;
+
+	re = *r++;
+#if SLANG_VERSION < 20000
+	if ((re->min_length > len)
+	    || (b != SLang_regexp_match (b, len, re)))
+	  continue;
+	match_len = re->end_matches[0];
+#else
+	if (b != (unsigned char *) SLregexp_match (re, (char *) b, len))
+	  continue;
+	(void) SLregexp_nth_match (re, 0, NULL, &match_len);
+#endif
+	return b + match_len;
+     }
+   return NULL;
+}
+
+/*}}}*/
 
 int _slrn_art_hide_quotes (Slrn_Article_Type *a, int reset) /*{{{*/
 {
    Slrn_Article_Line_Type *l, *last;
-   
+
    if (a == NULL)
      return -1;
 
    _slrn_art_unhide_quotes (a);
-   
+
    a->needs_sync = 1;
 
    if (!reset)
      Art_Hide_Quote_Level = Slrn_Quotes_Hidden_Mode;
-   
+
    l = a->lines;
    last = NULL;
 
@@ -147,13 +173,32 @@ int _slrn_art_hide_quotes (Slrn_Article_Type *a, int reset) /*{{{*/
 	  {
 	     if (Art_Hide_Quote_Level > 1)
 	       {
-		  if (QUOTE_LEVEL(l->flags) >= (unsigned int) Art_Hide_Quote_Level - 1)
+		  if (l->v.quote_level + 1 >= (unsigned int) Art_Hide_Quote_Level)
 		    l->flags |= HIDDEN_LINE;
 	       }
-	     else if (last != NULL)
-	       l->flags |= HIDDEN_LINE;
-	     
-	     last = l;
+	     /* Show first line of quoted material; try to pick one that
+	      * actually has text on it! */
+	     else
+	       {
+		  if (last != NULL)
+		    l->flags |= HIDDEN_LINE;
+		  else
+		    {
+		       last = l;
+		       if ((l->next != NULL) && (l->next->flags & QUOTE_LINE))
+			 {
+			    unsigned char *str, *line = (unsigned char*)l->buf;
+			    while (NULL != (str = is_matching_line
+					    (line, Slrn_Ignore_Quote_Regexp)))
+			      line = str;
+			    if (0 == *(slrn_skip_whitespace((char*)line)))
+			      {
+				 l->flags |= HIDDEN_LINE;
+				 last = NULL;
+			      }
+			 }
+		    }
+	       }
 	  }
 	else last = NULL;
 	l = l->next;
@@ -164,23 +209,22 @@ int _slrn_art_hide_quotes (Slrn_Article_Type *a, int reset) /*{{{*/
 
 /*}}}*/
 
-
 #if SLRN_HAS_SPOILERS
 void slrn_art_mark_spoilers (Slrn_Article_Type *a) /*{{{*/
 {
    Slrn_Article_Line_Type *l;
-   
+
    if ((Slrn_Spoiler_Char == 0) || (a == NULL))
      return;
 
    l = a->lines;
-   
+
    while ((l != NULL) && (l->flags & HEADER_LINE))
      l = l->next; /* skip header */
-   
-   while ((l != NULL) && (NULL == strchr (l->buf, 12)))
+
+   while ((l != NULL) && (NULL == slrn_strbyte (l->buf, 12)))
      l = l->next; /* skip to first formfeed */
-   
+
    while (l != NULL)
      {
 	l->flags |= SPOILER_LINE;
@@ -194,25 +238,37 @@ void slrn_art_mark_spoilers (Slrn_Article_Type *a) /*{{{*/
 static int try_supercite (Slrn_Article_Line_Type *l) /*{{{*/
 {
    Slrn_Article_Line_Type *last, *lsave;
+#if SLANG_VERSION < 20000
    static unsigned char compiled_pattern_buf[256];
-   static SLRegexp_Type re;
+   static SLRegexp_Type re_buf;
+   SLRegexp_Type *re;
+#else
+   static SLRegexp_Type *re = NULL;
+#endif
    unsigned char *b;
    int count;
    char name[32];
    unsigned int len;
+   unsigned int ofs;
    int ret;
 
-   re.pat = (unsigned char *) Super_Cite_Regexp;
-   re.buf = compiled_pattern_buf;
-   re.case_sensitive = 1;
-   re.buf_len = sizeof (compiled_pattern_buf);
-   
+#if SLANG_VERSION < 20000
+   re = &re_buf;
+   re->pat = (unsigned char *) Super_Cite_Regexp;
+   re->buf = compiled_pattern_buf;
+   re->case_sensitive = 1;
+   re->buf_len = sizeof (compiled_pattern_buf);
+   if ((*compiled_pattern_buf == 0) && SLang_regexp_compile (re))
+     return -1;
+#else
+   if ((re == NULL)
+       && (NULL == (re = SLregexp_compile (Super_Cite_Regexp, 0))))
+     return -1;
+#endif
+
    /* skip header --- I should look for Xnewsreader: gnus */
    while ((l != NULL) && (*l->buf != 0)) l = l->next;
-   
-   if ((*compiled_pattern_buf == 0) && SLang_regexp_compile (&re))
-     return -1;
-   
+
    /* look at the first 15 lines on first attempt.
     * After that, scan the whole buffer looking for more citations */
    count = 15;
@@ -224,7 +280,7 @@ static int try_supercite (Slrn_Article_Line_Type *l) /*{{{*/
 	  {
 	     if ((l->flags & QUOTE_LINE) == 0)
 	       {
-		  if (NULL != slrn_regexp_match (&re, l->buf))
+		  if (NULL != slrn_regexp_match (re, l->buf))
 		    {
 		       l->flags |= QUOTE_LINE;
 		       break;
@@ -233,26 +289,28 @@ static int try_supercite (Slrn_Article_Line_Type *l) /*{{{*/
 	     l = l->next;
 	     count--;
 	  }
-	
+
 	if ((l == NULL) || (count == 0)) return ret;
-	
+
 	/* Now find out what is used for citing. */
-	b = (unsigned char *) l->buf + re.beg_matches[1];
-	len = re.end_matches[1];
+	if (-1 == SLregexp_nth_match (re, 1, &ofs, &len))
+	  return ret;
+
+	b = (unsigned char *) l->buf + ofs;
 	if (len > sizeof (name) - 2) return ret;
-	
+
 	ret = 0;
 	strncpy (name, (char *) b, len); name[len] = 0;
-	
+
 	while (l != NULL)
 	  {
 	     unsigned char ch;
-	     
+
 	     b = (unsigned char *) l->buf;
 	     last = l;
 	     l = l->next;
 	     if (last->flags & QUOTE_LINE) continue;
-	     
+
 	     b = (unsigned char *) slrn_skip_whitespace ((char *) b);
 
 	     if (!strncmp ((char *) b, name, len)
@@ -260,7 +318,7 @@ static int try_supercite (Slrn_Article_Line_Type *l) /*{{{*/
 		     || (ch > 'z')))
 	       {
 		  last->flags |= QUOTE_LINE;
-		  
+
 		  while (l != NULL)
 		    {
 		       b = (unsigned char *) slrn_skip_whitespace (l->buf);
@@ -280,33 +338,11 @@ static int try_supercite (Slrn_Article_Line_Type *l) /*{{{*/
 
 /*}}}*/
 
-static unsigned char *is_matching_line (unsigned char *b, SLRegexp_Type **r) /*{{{*/
-{
-   unsigned int len;
-   len = strlen ((char *) b);
-
-   while (*r != NULL)
-     {
-	SLRegexp_Type *re;
-
-	re = *r++;
-	if ((re->min_length > len) 
-	    || (b != SLang_regexp_match (b, len, re)))
-	  continue;
-	
-	return b + re->end_matches[0];
-     }
-   return NULL;
-}
-
-/*}}}*/
-
-
 void slrn_art_mark_quotes (Slrn_Article_Type *a) /*{{{*/
 {
    Slrn_Article_Line_Type *l;
    unsigned char *b;
-   
+
    if (a == NULL)
      return;
 
@@ -314,9 +350,9 @@ void slrn_art_mark_quotes (Slrn_Article_Type *a) /*{{{*/
      {
 	/* return; */
      }
-   
+
    if (Slrn_Ignore_Quote_Regexp[0] == NULL) return;
-   
+
    /* skip header */
    l = a->lines;
    while ((l != NULL) && (l->flags == HEADER_LINE))
@@ -334,11 +370,10 @@ void slrn_art_mark_quotes (Slrn_Article_Type *a) /*{{{*/
 	  }
 	l->flags |= QUOTE_LINE;
 	level = 1;
-	while ((level < MAX_QUOTE_LEVELS) &&
-	       (NULL != (b = is_matching_line (b, Slrn_Ignore_Quote_Regexp))))
+	while (NULL != (b = is_matching_line (b, Slrn_Ignore_Quote_Regexp)))
 	  level++;
 
-	l->flags |= ((level-1) << QUOTE_LEVEL_SHIFT);
+	l->v.quote_level = level;
 	l = l->next;
      }
    a->is_modified = 1;
@@ -349,24 +384,25 @@ void slrn_art_mark_quotes (Slrn_Article_Type *a) /*{{{*/
 void slrn_art_mark_signature (Slrn_Article_Type *a) /*{{{*/
 {
    Slrn_Article_Line_Type *l;
-   
+
    if (a == NULL)
      return;
-   
+
    l = a->lines;
    if (l == NULL) return;
    /* go to end of article */
    while (l->next != NULL) l = l->next;
 
    /* skip back until a line matches the signature RegExp */
-   
+
    while ((l != NULL) && (0 == (l->flags & HEADER_LINE))
-	  && (NULL == is_matching_line ((unsigned char *) l->buf,
-					Slrn_Strip_Sig_Regexp)))
+	  && ((l->flags & VERBATIM_LINE) ||
+	      (NULL == is_matching_line ((unsigned char *) l->buf,
+					 Slrn_Strip_Sig_Regexp))))
      l = l->prev;
-   
+
    if ((l == NULL) || (l->flags & HEADER_LINE)) return;
-   
+
    while (l != NULL)
      {
         l->flags |= SIGNATURE_LINE;
@@ -397,7 +433,7 @@ static Slrn_Article_Line_Type *unwrap_line (Slrn_Article_Type *a, /*{{{*/
 
    ll = l->prev;
    b = ll->buf;
-   do 
+   do
      {
 	b += strlen (b);
 	/* skip the space at beginning of the wrapped line: */
@@ -409,7 +445,7 @@ static Slrn_Article_Line_Type *unwrap_line (Slrn_Article_Type *a, /*{{{*/
 	l = next;
      }
    while ((l != NULL) && (l->flags & WRAPPED_LINE));
-	     
+
    ll->next = l;
    if (l != NULL) l->prev = ll;
    return l;
@@ -422,7 +458,7 @@ static Slrn_Article_Line_Type *unwrap_line (Slrn_Article_Type *a, /*{{{*/
 int _slrn_art_unwrap_article (Slrn_Article_Type *a) /*{{{*/
 {
    Slrn_Article_Line_Type *l;
-   
+
    if (a == NULL)
      return -1;
 
@@ -444,10 +480,13 @@ int _slrn_art_unwrap_article (Slrn_Article_Type *a) /*{{{*/
 
 int _slrn_art_wrap_article (Slrn_Article_Type *a) /*{{{*/
 {
-   unsigned int len;
    unsigned char *buf, ch;
    Slrn_Article_Line_Type *l;
    unsigned int wrap_mode = Slrn_Wrap_Mode;
+   int wrap_width = Slrn_Wrap_Width;
+
+   if (wrap_width < 5) wrap_width = 5; /* arbitrary */
+   else if (wrap_width > SLtt_Screen_Cols) wrap_width = SLtt_Screen_Cols;
 
    if (a == NULL)
      return -1;
@@ -460,11 +499,15 @@ int _slrn_art_wrap_article (Slrn_Article_Type *a) /*{{{*/
 
    l = a->lines;
 
+#if SLANG_VERSION >= 20000
+   SLsmg_gotorc(0,0); /* used by Slsmg_strbytes later on */
+#endif
+
    while (l != NULL)
      {
 	unsigned char header_char_delimiter = 0;
 
-	if (l->flags & HEADER_LINE) 
+	if (l->flags & HEADER_LINE)
 	  {
 	     if ((wrap_mode & 1) == 0)
 	       {
@@ -472,11 +515,9 @@ int _slrn_art_wrap_article (Slrn_Article_Type *a) /*{{{*/
 		  continue;
 	       }
 
-	     if (0 == slrn_case_strncmp ((unsigned char *)"Path: ",
-					 (unsigned char *)l->buf, 6))
+	     if (0 == slrn_case_strncmp ("Path: ", l->buf, 6))
 	       header_char_delimiter = '!';
-	     else if (0 == slrn_case_strncmp ((unsigned char *) "Newsgroups: ",
-					      (unsigned char *)l->buf, 12))
+	     else if (0 == slrn_case_strncmp ( "Newsgroups: ", l->buf, 12))
 	       header_char_delimiter = ',';
 	  }
 	else if (l->flags & QUOTE_LINE)
@@ -488,44 +529,34 @@ int _slrn_art_wrap_article (Slrn_Article_Type *a) /*{{{*/
 	       }
 	  }
 
-	len = 0;
 	buf = (unsigned char *) l->buf;
 	ch = *buf;
 	while (ch != 0)
 	  {
-	     if ((ch == '\t') && (SLsmg_Tab_Width > 0))
-	       {
-		  len += SLsmg_Tab_Width;
-		  len -= len % SLsmg_Tab_Width;
-	       }
-	     else if (((ch >= ' ') && (ch < 127))
-		      || (ch >= (unsigned char) SLsmg_Display_Eight_Bit))
-	       len++;
-	     else
-	       {
-		  len += 2;
-		  if (ch & 0x80) len++;
-	       }
-
-	     if (len > (unsigned int) SLtt_Screen_Cols)
+	     unsigned int bytes = SLsmg_strbytes (buf, buf+strlen((char *)buf),
+						  (unsigned int) wrap_width);
+	     buf += bytes;
+	     while ((*buf == ' ') || (*buf == '\t'))
+	       buf++;
+	     if (*buf) /* we need to wrap */
 	       {
 		  Slrn_Article_Line_Type *new_l;
 		  unsigned char *buf0, *lbuf;
 
-		  /* Try to break the line on a word boundary.  
+		  /* Try to break the line on a word boundary.
 		   * For now, I will only break on space characters.
 		   */
 		  buf0 = buf;
 		  lbuf = (unsigned char *) l->buf;
 
 		  lbuf += 1;	       /* avoid space at beg of line */
-		  
+
 		  if (Slrn_Wrap_Method == 0 || Slrn_Wrap_Method == 2)
 		    {
 		       while (buf0 > lbuf)
 			 {
 			    if ((*buf0 == ' ') || (*buf0 == '\t')
-				|| (header_char_delimiter 
+				|| (header_char_delimiter
 				    && (*buf0 == header_char_delimiter)))
 			      {
 				 buf = buf0;
@@ -533,19 +564,19 @@ int _slrn_art_wrap_article (Slrn_Article_Type *a) /*{{{*/
 			      }
 			    buf0--;
 			 }
-		       
+
 		       if (buf0 == lbuf)
 			 {
 			    if (Slrn_Wrap_Method == 0)
 			      {
 		       /* Could not find a place to break the line.  Ok, so
-			* we will not break this.  Perhaps it is a URL.  
+			* we will not break this.  Perhaps it is a URL.
 			* If not, it is a long word and who cares about it.
 			*/
 				 while (((ch = *buf) != 0)
 					&& (ch != ' ') && (ch != '\t'))
 				   buf++;
-				 
+
 				 if (ch == 0)
 				   continue;
 			      }
@@ -555,7 +586,7 @@ int _slrn_art_wrap_article (Slrn_Article_Type *a) /*{{{*/
 			    }
 			 }
 		    }
-		  
+
 		  /* Start wrapped lines with a space.  To do this, I will
 		   * _temporally_ modify the previous character for the purpose
 		   * of creating the new space.
@@ -563,11 +594,11 @@ int _slrn_art_wrap_article (Slrn_Article_Type *a) /*{{{*/
 		  buf--;
 		  ch = *buf;
 		  *buf = ' ';
-		  
+
 		  new_l = (Slrn_Article_Line_Type *) slrn_malloc (sizeof (Slrn_Article_Line_Type), 1, 1);
 		  if (new_l == NULL)
 		    return -1;
-		  
+
 		  if (NULL == (new_l->buf = slrn_strmalloc ((char *)buf, 1)))
 		    {
 		       slrn_free ((char *) new_l);
@@ -576,20 +607,25 @@ int _slrn_art_wrap_article (Slrn_Article_Type *a) /*{{{*/
 
 		  *buf++ = ch;
 		  *buf = 0;
-		  
+
 		  new_l->next = l->next;
 		  new_l->prev = l;
 		  l->next = new_l;
 		  if (new_l->next != NULL) new_l->next->prev = new_l;
 
 		  new_l->flags = l->flags | WRAPPED_LINE;
+
+		  if (new_l->flags & QUOTE_LINE)
+		    new_l->v.quote_level = l->v.quote_level;
+
 		  l = new_l;
 		  buf = (unsigned char *) new_l->buf;
-		  len = 0;
 		  a->is_wrapped = 1;
 	       }
+#if SLANG_VERSION < 20000
 	     else buf++;
-	     
+#endif
+
 	     ch = *buf;
 	  }
 	l = l->next;
@@ -599,7 +635,6 @@ int _slrn_art_wrap_article (Slrn_Article_Type *a) /*{{{*/
 
 /*}}}*/
 
-
 static int is_blank_line (unsigned char *b) /*{{{*/
 {
    b = (unsigned char *) slrn_skip_whitespace ((char *) b);
@@ -628,16 +663,16 @@ void _slrn_art_skip_quoted_text (Slrn_Article_Type *a) /*{{{*/
 	  }
 	l = l->next;
      }
-   
+
    /* Now we are either at the end of the buffer or on a quote line. Skip
     * past other quote lines.
     */
-   
+
    if (l == NULL)
      return;
-   
+
    l = l->next;
-   
+
    while (l != NULL)
      {
 	if (l->flags & HIDDEN_LINE)
@@ -687,13 +722,13 @@ int _slrn_art_skip_digest_forward (Slrn_Article_Type *a) /*{{{*/
 	  {
 	     char ch;
 	     char *buf;
-	     
+
 	     if ((l->flags & HIDDEN_LINE) || (l->flags & HEADER_LINE))
 	       {
 		  l = l->next;
 		  continue;
 	       }
-	     
+
 	     buf = l->buf;
 	     if (num_passes == 0)
 	       {
@@ -713,7 +748,7 @@ int _slrn_art_skip_digest_forward (Slrn_Article_Type *a) /*{{{*/
 		       continue;
 		    }
 	       }
-	     
+
 	     a->cline = l;
 	     a->needs_sync = 1;
 	     return 0;
@@ -732,14 +767,14 @@ char *slrn_art_extract_header (char *hdr, unsigned int len) /*{{{*/
 
    if (a == NULL)
      return NULL;
-   
+
    l = a->lines;
 
    while ((l != NULL)
 	  && (*l->buf != 0))
      {
-	if (0 == slrn_case_strncmp ((unsigned char *) hdr,
-				    (unsigned char *) l->buf, len))
+	if (0 == slrn_case_strncmp ( hdr,
+				     l->buf, len))
 	  {
 	     char *result;
 
@@ -750,8 +785,8 @@ char *slrn_art_extract_header (char *hdr, unsigned int len) /*{{{*/
 	       }
 
 	     /* Return the data after the colon */
-	     result = slrn_strchr (l->buf, ':');
-	     if (result == NULL) 
+	     result = slrn_strbyte (l->buf, ':');
+	     if (result == NULL)
 	       result = l->buf + len;
 	     else result += 1;
 
@@ -776,9 +811,7 @@ Visible_Header_Type;
 
 static Visible_Header_Type *Visible_Headers;
 
-#if SLRN_HAS_SLANG
 char *Slrn_Visible_Headers_String;     /* for interpreter */
-#endif
 
 static void free_visible_header_list (void) /*{{{*/
 {
@@ -790,32 +823,28 @@ static void free_visible_header_list (void) /*{{{*/
 	SLfree ((char *) Visible_Headers);
 	Visible_Headers = next;
      }
-#if SLRN_HAS_SLANG
    SLang_free_slstring (Slrn_Visible_Headers_String);
-#endif
+   Slrn_Visible_Headers_String = NULL;
 }
 
 /*}}}*/
 
-
 int slrn_set_visible_headers (char *headers) /*{{{*/
 {
    char buf[256];
    unsigned int nth;
 
    free_visible_header_list ();
-   
-#if SLRN_HAS_SLANG
+
    Slrn_Visible_Headers_String = SLang_create_slstring (headers);
    if (Slrn_Visible_Headers_String == NULL)
      return -1;
-#endif
 
    nth = 0;
    while (-1 != SLextract_list_element (headers, nth, ',', buf, sizeof(buf)))
      {
 	Visible_Header_Type *next;
-	
+
 	next = (Visible_Header_Type *) SLmalloc (sizeof (Visible_Header_Type));
 	if (next == NULL)
 	  return -1;
@@ -828,7 +857,7 @@ int slrn_set_visible_headers (char *headers) /*{{{*/
 	next->len = strlen (buf);
 	next->next = Visible_Headers;
 	Visible_Headers = next;
-	
+
 	nth++;
      }
    if ((Slrn_Current_Article != NULL) &&
@@ -843,21 +872,21 @@ void _slrn_art_hide_headers (Slrn_Article_Type *a) /*{{{*/
 {
    Slrn_Article_Line_Type *l;
    char ch;
-   
+
    if (a == NULL)
      return;
 
    _slrn_art_unhide_headers (a);
 
    a->needs_sync = 1;
-   
+
    l = a->lines;
 
    while ((l != NULL) && (l->flags & HEADER_LINE))
      {
 	int hide_header;
 	Visible_Header_Type *v;
-	
+
 	ch = l->buf[0];
 	ch |= 0x20;
 
@@ -868,16 +897,16 @@ void _slrn_art_hide_headers (Slrn_Article_Type *a) /*{{{*/
 	  {
 	     int hide = (v->header[0] == '!') ? 1 : 0;
 	     char chv = (0x20 | v->header[hide]);
-	     
+
 	     if ((chv == ch)
-		 && (0 == slrn_case_strncmp ((unsigned char *)l->buf,
-					     (unsigned char *)v->header + hide,
+		 && (0 == slrn_case_strncmp (l->buf,
+					     v->header + hide,
 					     (v->len) - hide)))
 	       {
 		  hide_header = hide;
 		  break;
 	       }
-	     
+
 	     v = v->next;
 	  }
 
@@ -885,7 +914,7 @@ void _slrn_art_hide_headers (Slrn_Article_Type *a) /*{{{*/
 	  {
 	     if (hide_header)
 	       l->flags |= HIDDEN_LINE;
-	     
+
 	     l = l->next;
 	  }
 	while ((l != NULL) && ((*l->buf == ' ') || (*l->buf == '\t')));
@@ -910,10 +939,10 @@ int _slrn_art_unfold_header_lines (Slrn_Article_Type *a) /*{{{*/
 {
    Slrn_Article_Line_Type *l;
    char ch;
-   
+
    if (a == NULL)
      return -1;
-   
+
    l = a->lines->next;
    a->is_modified = 1;
    a->needs_sync = 1;
@@ -926,20 +955,20 @@ int _slrn_art_unfold_header_lines (Slrn_Article_Type *a) /*{{{*/
 	     unsigned int len0, len1;
 	     Slrn_Article_Line_Type *prev;
 	     char *new_buf;
-	     
+
 	     l->buf[0] = ' ';
-	     
+
 	     prev = l->prev;
-	     
+
 	     len0 = strlen (prev->buf);
 	     len1 = len0 + strlen (l->buf) + 1;
-	     
+
 	     new_buf = slrn_realloc (prev->buf, len1, 1);
 	     if (new_buf == NULL)
 	       return -1;
 
 	     prev->buf = new_buf;
-	     
+
 	     strcpy (new_buf + len0, l->buf); /* safe */
 	     prev->next = l->next;
 	     if (l->next != NULL) l->next->prev = prev;
@@ -947,7 +976,7 @@ int _slrn_art_unfold_header_lines (Slrn_Article_Type *a) /*{{{*/
 	     slrn_free ((char *) l);
 	     l = prev;
 	  }
-	
+
 	l = l->next;
      }
    return 0;
@@ -958,7 +987,7 @@ int _slrn_art_unfold_header_lines (Slrn_Article_Type *a) /*{{{*/
 void slrn_mark_header_lines (Slrn_Article_Type *a) /*{{{*/
 {
    Slrn_Article_Line_Type *l;
-   
+
    if (a == NULL)
      return;
    l = a->lines;
@@ -995,18 +1024,46 @@ void _slrn_art_unhide_signature (Slrn_Article_Type *a) /*{{{*/
 /*}}}*/
 
 void slrn_art_mark_pgp_signature (Slrn_Article_Type *a) /*{{{*/
-{	
+{
    Slrn_Article_Line_Type *l;
 
    if (a == NULL)
      return;
-   
+
    l = a->lines;
    while (l != NULL)
      {
 	Slrn_Article_Line_Type *l0;
 	int count;
 
+	if ((l->buf[0] == '-')
+	    && !strcmp (l->buf, "-----BEGIN PGP SIGNED MESSAGE-----"))
+	  {
+	     l->flags |= PGP_SIGNATURE_LINE;
+	     l->flags &= ~QUOTE_LINE;
+
+	     if ((NULL != (l = l->next)) &&
+		 !strncmp (l->buf, "Hash: ", 6))
+	       {
+		  /* catch the `Hash: ... ' line */
+		  l->flags |= PGP_SIGNATURE_LINE;
+		  l->flags &= ~QUOTE_LINE;
+
+		  l = l->next;
+	       }
+	     if ((NULL != l) &&
+		 !strncmp (l->buf, "NotDashEscaped: ", 16))
+	       {
+		  /* the optional "NotDashEscaped: ..." line */
+		  l->flags |= PGP_SIGNATURE_LINE;
+		  l->flags &= ~QUOTE_LINE;
+
+		  l = l->next;
+	       }
+
+	     continue;
+	  }
+
 	if ((l->buf[0] != '-')
 	    || strcmp (l->buf, "-----BEGIN PGP SIGNATURE-----"))
 	  {
@@ -1015,8 +1072,8 @@ void slrn_art_mark_pgp_signature (Slrn_Article_Type *a) /*{{{*/
 	  }
 	l0 = l;
 	l = l->next;
-	
-	count = 20;
+
+	count = 256;		       /* arbitrary */
 	while ((l != NULL) && count)
 	  {
 	     if ((l->buf[0] != '-')
@@ -1026,14 +1083,14 @@ void slrn_art_mark_pgp_signature (Slrn_Article_Type *a) /*{{{*/
 		  l = l->next;
 		  continue;
 	       }
-	     
+
 	     l0->flags |= PGP_SIGNATURE_LINE;
-	     l0->flags &= ~QUOTE_LINE;
+	     l0->flags &= ~(QUOTE_LINE|SIGNATURE_LINE);
 	     do
 	       {
 		  l0 = l0->next;
 		  l0->flags |= PGP_SIGNATURE_LINE;
-		  l0->flags &= ~QUOTE_LINE;
+		  l0->flags &= ~(QUOTE_LINE|SIGNATURE_LINE);
 	       }
 	     while (l0 != l);
 	     return;
@@ -1066,7 +1123,6 @@ void _slrn_art_unhide_pgp_signature (Slrn_Article_Type *a) /*{{{*/
 
 /*}}}*/
 
-#if SLRN_HAS_VERBATIM_MARKS
 void slrn_art_mark_verbatim (Slrn_Article_Type *a)
 {
    Slrn_Article_Line_Type *l;
@@ -1076,7 +1132,7 @@ void slrn_art_mark_verbatim (Slrn_Article_Type *a)
 
    if (a == NULL)
      return;
-   
+
    von = "#v+";
    voff = "#v-";
 
@@ -1115,7 +1171,6 @@ void slrn_art_mark_verbatim (Slrn_Article_Type *a)
      }
 }
 
-
 void _slrn_art_unhide_verbatim (Slrn_Article_Type *a) /*{{{*/
 {
    if (a == NULL)
@@ -1137,5 +1192,3 @@ void _slrn_art_hide_verbatim (Slrn_Article_Type *a) /*{{{*/
 }
 
 /*}}}*/
-
-#endif
