@@ -1,8 +1,8 @@
 /*
  This file is part of SLRN.
 
- Copyright (c) 1994, 1999 John E. Davis <davis@space.mit.edu>
- Copyright (c) 2001, 2002 Thomas Schultz <tststs@gmx.de>
+ Copyright (c) 1994, 1999, 2007-2012 John E. Davis <jed@jedsoft.org>
+ Copyright (c) 2001-2006 Thomas Schultz <tststs@gmx.de>
 
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
@@ -19,22 +19,12 @@
  59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-/* Keep this separate from autoconf's settings for now */
-#define MY_VERSION "0.9.7.4"
-#define SLRN_VERSION_NUMBER 90704
-/* #define PATCH_LEVEL "1" */
-#define SLRN_RELEASE_DATE "2002-03-13"
-
-#ifdef PATCH_LEVEL
-# define SLRN_VERSION (MY_VERSION "pl" PATCH_LEVEL)
-#else  
-# define SLRN_VERSION (MY_VERSION)
-#endif
+#define SLRN_VERSION_STRING "1.0.0"
+#define SLRN_VERSION 10000
 
 #ifndef SLRNPULL_CODE
-extern char *Slrn_Version;
-extern int Slrn_Version_Number;
-extern char *Slrn_Date;
+extern char *Slrn_Version_String;
+extern int Slrn_Version;
 extern char *slrn_get_os_name (void);
-extern void slrn_show_version (void);
+extern void slrn_show_version (FILE *);
 #endif
